

import copy
import logging
import os
import subprocess as sp
import sys

import fantail
from leip import set_local_config


lg = logging.getLogger(__name__)


def get_tool_conf(app, name, version='default'):

    data = copy.copy(app.conf['group.default'])
    tool_data = copy.copy(app.conf.get('app.{}'.format(name), fantail.Fantail()))
    group = tool_data.get('group')

    if not group is None:
        group_data = app.conf['group.{}'.format(group)]
        if group_data:
            data.update(group_data)

    data.update(tool_data)

    if version is 'default':
        version = tool_data.get('default_version', None)

    if (not version is None) and (not version in tool_data['versions']):
        candidates = []
        for v in tool_data['versions']:
            fullv =  tool_data['versions'][v]['version']
            if v in fullv:
                candidates.append(v)

    if not version is None:
        version_data = tool_data['versions.{}'.format(version)]
        data.update(version_data)
        data['version_key'] = version

    return data


def is_kea(fname):

    with open(fname) as F:
        start = F.read(1000)

    fline = start.strip().split("\n")[0]
    if not fline.startswith('#!'):
        lg.debug(" - not a shell script - not kea")
        return False

    if not 'python' in fline:
        lg.debug(" - not a python script - not kea")
        return False

    if 'load_entry_point' in start and \
            'Kea==' in start:
        lg.debug(" - looks like a link to the kea entry point script - kea")
        return True

    if 'import Kea' in start or \
            'from Kea import' in start:
        lg.debug(" - looks like custom Kea script - kea")
        return True

    lg.debug(" - does not look like a kea script")
    return False


def find_executable(name):

    # check if this is a single executable:
    if os.path.isfile(name) and os.access(name, os.X_OK):
        executable = name
        name = os.path.basename(executable)
        yield os.path.abspath(executable)

    else:

        # no? try to use the 'which' tool

        # no '/' allowed anymore
        if '/' in name:
            raise IOError(name)

        P = sp.Popen(['which', '-a', name], stdout=sp.PIPE)

        out, err = P.communicate()

        for line in out.strip().split("\n"):
            lg.debug("check %s", line)
            if not is_kea(line):
                lg.debug("%s is not a kea file", line)

                yield os.path.abspath(line)


def create_kea_link(app, name):
    """
    """
    base = app.conf['bin_path']
    linkpath = os.path.expanduser(os.path.join(base, name))
    lg.debug("checking: %s", linkpath)
    if os.path.lexists(linkpath):
        lg.debug("path exists: %s", linkpath)
        os.unlink(linkpath)


    keapath = sys.argv[0]
    lg.info("creating link from %s", linkpath)
    lg.info(" to: %s", keapath)
    os.symlink(keapath, linkpath)

def register_executable(app, name, executable, version, is_default=None):
    """
    Register an executable
    """

    allversions = list('abcdefghijklmnopqrstuvwxyz123456789')

    is_first_version = True

    version_key = 'a'

    if app.conf.has_key('app.{}.versions'.format(name)):
        is_first_version = False
        for k in app.conf['app.{}.versions'.format(name)]:
            vinf = app.conf['app.{}.versions.{}'\
                            .format(name, k)]
            if vinf['executable'] == executable:
                lg.warning("Executable is already registered - overwriting")
                version_key = k
                break

            #registered - we do not want to use this key
            allversions.remove(k)

        version_key = allversions[0]

    if is_default == False:
        if is_first_version:
            lg.debug("First version of %s - setting to default", name)
            is_default = True
        else:
            lg.debug("Other version of %s present - not setting default", name)
            is_default = False

    lg.warning("register %s - %s - %s - %s", name, executable,
             version_key, version)

    if is_default:
        lg.warning("Set version %s as default", version_key)
        set_local_config(app, 'app.{}.default_version'.format(name),
                         version_key)

    basekey = 'app.{}.versions.{}'.format(name, version_key)
    lg.debug("register to: %s", basekey)

    set_local_config(app, '{}.executable'.format(basekey), executable)
    set_local_config(app, '{}.version'.format(basekey), version)

    create_kea_link(app, name)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# CAVEAT UTILITOR
#
# This file was automatically generated by Grako.
#
#    https://pypi.python.org/pypi/grako/
#
# Any changes you make to it will be overwritten the next time
# the file is generated.


from __future__ import print_function, division, absolute_import, unicode_literals
from grako.parsing import graken, Parser


__version__ = (2014, 12, 11, 12, 59, 52, 3)

__all__ = [
    'samtoolsParser',
    'samtoolsSemantics',
    'main'
]


class samtoolsParser(Parser):
    def __init__(self, whitespace=None, nameguard=True, **kwargs):
        super(samtoolsParser, self).__init__(
            whitespace=whitespace,
            nameguard=nameguard,
            **kwargs
        )

    @graken()
    def _start_(self):
        self._token('samtools')
        self._command_()
        self._check_eof()

    @graken()
    def _command_(self):
        with self._choice():
            with self._option():
                self._samtools_view_()
            with self._option():
                self._samtools_merge_()
            with self._option():
                self._samtools_flagstat_()
            with self._option():
                self._samtools_index_()
            with self._option():
                self._samtools_sort_()
            with self._option():
                self._samtools_empty_()
            self._error('no available options')

    @graken()
    def _samtools_empty_(self):
        pass

    @graken()
    def _samtools_view_(self):
        self._token('view')
        self.ast['command'] = self.last_node

        def block1():
            self._view_option_()
            self.ast['options'] = self.last_node
        self._closure(block1)

        def block4():
            self._bamsamfile_()
        self._closure(block4)
        self.ast['input'] = self.last_node

        self.ast._define(
            ['command', 'options', 'input'],
            []
        )

    @graken()
    def _samtools_merge_(self):
        self._token('merge')
        self.ast['command'] = self.last_node

        def block1():
            self._merge_option_()
            self.ast['options'] = self.last_node
        self._closure(block1)
        self._bamsamfile_()
        self.ast['output'] = self.last_node

        def block5():
            self._bamsamfile_()
        self._positive_closure(block5)

        self.ast['input'] = self.last_node

        self.ast._define(
            ['command', 'options', 'output', 'input'],
            []
        )

    @graken()
    def _samtools_flagstat_(self):
        self._token('flagstat')
        self.ast['command'] = self.last_node
        self._bamsamfile_()
        self.ast['input'] = self.last_node

        self.ast._define(
            ['command', 'input'],
            []
        )

    @graken()
    def _samtools_index_(self):
        self._token('index')
        self.ast['command'] = self.last_node
        self._bamsamfile_()
        self.ast['input'] = self.last_node

        def block3():
            self._anyfile_()
        self._closure(block3)
        self.ast['output'] = self.last_node

        self.ast._define(
            ['command', 'input', 'output'],
            []
        )

    @graken()
    def _samtools_sort_(self):
        self._token('sort')
        self.ast['command'] = self.last_node

        def block1():
            self._sort_option_()
            self.ast['options'] = self.last_node
        self._closure(block1)
        self._bamsamfile_()
        self.ast['input'] = self.last_node

        def block5():
            self._anyfile_()
        self._closure(block5)
        self.ast['output_prefix'] = self.last_node

        self.ast._define(
            ['command', 'options', 'input', 'output_prefix'],
            []
        )

    @graken()
    def _view_option_(self):

        def block0():
            with self._choice():
                with self._option():
                    self._view_argoption_()
                with self._option():
                    self._view_noargoption_()
                self._error('no available options')
        self._closure(block0)

    @graken()
    def _view_argoption_(self):
        self._view_argoptionflag_()
        self._argoptionvalue_()

    @graken()
    def _view_argoptionflag_(self):
        self._pattern(r'-[@LtToRfFqlrs]')

    @graken()
    def _view_noargoption_(self):
        self._pattern(r'-[bhHSu1xXcB]+')

    @graken()
    def _sort_option_(self):

        def block0():
            with self._choice():
                with self._option():
                    self._sort_argoption_()
                with self._option():
                    self._sort_noargoption_()
                self._error('no available options')
        self._closure(block0)

    @graken()
    def _sort_argoption_(self):
        self._sort_argoptionflag_()
        self._argoptionvalue_()

    @graken()
    def _sort_argoptionflag_(self):
        self._pattern(r'-[l@m]')

    @graken()
    def _sort_noargoption_(self):
        self._pattern(r'-[nfo]+')

    @graken()
    def _merge_option_(self):

        def block0():
            with self._choice():
                with self._option():
                    self._merge_argoption_()
                with self._option():
                    self._merge_noargoption_()
                self._error('no available options')
        self._closure(block0)

    @graken()
    def _merge_argoption_(self):
        self._merge_argoptionflag_()
        self._argoptionvalue_()

    @graken()
    def _merge_argoptionflag_(self):
        self._pattern(r'-[l@Rh]+')

    @graken()
    def _merge_noargoption_(self):
        self._pattern(r'-[nruf1]+')

    @graken()
    def _argoptionvalue_(self):
        self._pattern(r'[^ ]+')

    @graken()
    def _bamsamfile_(self):
        self._pattern(r'[^ ]+\.[bs]am')

    @graken()
    def _anyfile_(self):
        self._pattern(r'[^ ]+')


class samtoolsSemantics(object):
    def start(self, ast):
        return ast

    def command(self, ast):
        return ast

    def samtools_empty(self, ast):
        return ast

    def samtools_view(self, ast):
        return ast

    def samtools_merge(self, ast):
        return ast

    def samtools_flagstat(self, ast):
        return ast

    def samtools_index(self, ast):
        return ast

    def samtools_sort(self, ast):
        return ast

    def view_option(self, ast):
        return ast

    def view_argoption(self, ast):
        return ast

    def view_argoptionflag(self, ast):
        return ast

    def view_noargoption(self, ast):
        return ast

    def sort_option(self, ast):
        return ast

    def sort_argoption(self, ast):
        return ast

    def sort_argoptionflag(self, ast):
        return ast

    def sort_noargoption(self, ast):
        return ast

    def merge_option(self, ast):
        return ast

    def merge_argoption(self, ast):
        return ast

    def merge_argoptionflag(self, ast):
        return ast

    def merge_noargoption(self, ast):
        return ast

    def argoptionvalue(self, ast):
        return ast

    def bamsamfile(self, ast):
        return ast

    def anyfile(self, ast):
        return ast


def main(filename, startrule, trace=False, whitespace=None, nameguard=None):
    import json
    with open(filename) as f:
        text = f.read()
    parser = samtoolsParser(parseinfo=False)
    ast = parser.parse(
        text,
        startrule,
        filename=filename,
        trace=trace,
        whitespace=whitespace,
        nameguard=nameguard)
    print('AST:')
    print(ast)
    print()
    print('JSON:')
    print(json.dumps(ast, indent=2))
    print()

if __name__ == '__main__':
    import argparse
    import string
    import sys

    class ListRules(argparse.Action):
        def __call__(self, parser, namespace, values, option_string):
            print('Rules:')
            for r in samtoolsParser.rule_list():
                print(r)
            print()
            sys.exit(0)

    parser = argparse.ArgumentParser(description="Simple parser for samtools.")
    parser.add_argument('-l', '--list', action=ListRules, nargs=0,
                        help="list all rules and exit")
    parser.add_argument('-n', '--no-nameguard', action='store_true',
                        dest='no_nameguard',
                        help="disable the 'nameguard' feature")
    parser.add_argument('-t', '--trace', action='store_true',
                        help="output trace information")
    parser.add_argument('-w', '--whitespace', type=str, default=string.whitespace,
                        help="whitespace specification")
    parser.add_argument('file', metavar="FILE", help="the input file to parse")
    parser.add_argument('startrule', metavar="STARTRULE",
                        help="the start rule for parsing")
    args = parser.parse_args()

    main(
        args.file,
        args.startrule,
        trace=args.trace,
        whitespace=args.whitespace,
        nameguard=not args.no_nameguard
    )




#!/usr/bin/env python

import glob
import os
import sys
import tempfile
import subprocess as sp

if len(sys.argv) > 1 and sys.argv[1] == '-v':
    VERBOSE = True
    sys.argv = sys.argv[2:]
else:
    VERBOSE = False

if len(sys.argv) > 1:
    select = " ".join(sys.argv[1:])
else:
    select = ""

for ymlf in glob.glob('*.yaml'):
    ebnffile = ymlf.replace('.yaml', '.ebnf')
    if os.path.exists(ebnffile):
        ebnftime = os.stat(ebnffile).st_mtime
        ymltime = os.stat(ymlf).st_mtime
        if ebnftime > ymltime:
            print "Skipping %s (not changed)" % ymlf
            continue

    print "convert %s -> %s" % (ymlf, ebnffile)
    cl = './ebnfgenerator.py {} > {}'.format(ymlf, ebnffile)
    os.system(cl)

for ebnf in glob.glob('*.ebnf'):
    pyfile = ebnf.replace('.ebnf', '.py')
    testfile = ebnf.replace('.ebnf', '.test')

    if os.path.exists(pyfile):
        pyfiletime = os.stat(pyfile).st_mtime
        ebnffiletime = os.stat(ebnf).st_mtime
        testtime =  os.stat(testfile).st_mtime
        if pyfiletime >= ebnffiletime and \
           pyfiletime >= testtime:
            print("skipping %s (not changed" % ebnf)
            continue

    cl = "grako -w ' ' {} -o {}".format(ebnf, pyfile)
    P = sp.Popen(cl, shell=True)
    P.communicate()
    if P.returncode != 0:
        print "compilation fail", ebnf
        os.unlink(pyfile)
        exit(-1)

    
    testsnippet = ""
    def test(pyfile, snip):

        if select and (not select in snip):
            return
        
        snip = snip.strip()
        
        if not snip: return

        if snip[0] == '-':
            shouldfail = True
            snip = snip[1:].strip()
        else:
            shouldfail = False
            
        tf = tempfile.NamedTemporaryFile('w', delete=False)
        tf.write(snip)
        tf.close()

        cl = "python {} {} start".format(pyfile, tf.name)
        P = sp.Popen(cl, shell=True, stdout=sp.PIPE, stderr=sp.PIPE)
        out, err = P.communicate()
        rc = P.returncode

        if shouldfail and rc != 0:
            print "OK (FAIL)", snip
            if VERBOSE:
                print out
            return True
        elif (not shouldfail) and rc == 0:
            print 'OK', snip
            if VERBOSE:
                print out
            return True
        elif shouldfail and rc == 0:
            print "NOT OK (did not fail)", snip
            print out
            print err
            exit(-1)
        else:
            print "NOT OK", snip
            print out
            print err
            exit(-1)
        
    with open(testfile) as F:
        for line in F:
            line = line.strip()
            if not line:
                test(pyfile, testsnippet)
                testsnippet = ""
            else:
                testsnippet += " " + line
                
    test(pyfile, testsnippet)
        





from collections import deque
import copy
from datetime import datetime
import fcntl
import hashlib
import logging
from multiprocessing.dummy import Pool as ThreadPool
import os
import subprocess as sp
import sys
import time


import psutil

import leip

lg = logging.getLogger(__name__)


@leip.hook('pre_argparse')
def main_arg_define(app):
    if app.executor == 'simple':
        app.parser.add_argument('-j', '--threads', help='no threads to use', type=int)


#thanks: https://gist.github.com/sebclaeys/1232088
def non_block_read(stream, chunk_size=10000):
    #print('xxx', type(stream))
    fd = stream.fileno()
    fl = fcntl.fcntl(fd, fcntl.F_GETFL)
    fcntl.fcntl(fd, fcntl.F_SETFL, fl | os.O_NONBLOCK)
    try:
        return stream.read()
    except:
        return ""

    
def streamer(src, tar, dq, hsh=None):
    """
    :param src: input stream
    :param tar: target stream
    :param dq: deque object keeping a tail of chunks for this stream
    :param hsh: hash object to calculate a checksum 
    """
    d = non_block_read(src)
    if d is None:
        return 0
    if hsh:
        hsh.update(d)
    
    dq.append(d.decode('utf-8'))
    d_len = len(d)
    tar.write(d) #d.encode('utf-8'))
    return d_len

    
def get_deferred_cl(info):
    dcl = ['kea']
    if info['stdout_file']:
        dcl.extend(['-o', info['stdout_file']])
    if info['stderr_file']:
        dcl.extend(['-e', info['stderr_file']])
    dcl.extend(info['cl'])
    return cl

def store_process_info(info):
    psu = info.get('psutil_process')
    if not psu: return
    try:
        info['ps_nice'] = psu.nice()
        info['ps_num_fds'] = psu.num_fds()
        info['ps_threads'] = psu.num_threads()
        cputime = psu.cpu_times()
        info['ps_cputime_user'] = cputime.user 
        info['ps_cputime_system'] = cputime.system
        meminfo = psu.memory_info()
        for f in meminfo._fields:
            #info['ps_meminfo_{}'.format(f)] = getattr(meminfo, f)
            info['ps_meminfo_max_{}'.format(f)] = \
                    max(getattr(meminfo, f),
                        info.get('ps_meminfo_max_{}'.format(f), 0))

        try:
            ioc = psu.io_counters()
            info['ps_io_read_count'] = ioc.read_count
        except AttributeError:
            #may not have iocounters (osx)
            pass

    except psutil.NoSuchProcess, psutil.AccessDenied:
        #process went away??
        return
            
def simple_runner(info, defer_run=False):
    """
    Defer run executes the run with the current executor, but with
    the Kea executable so that all kea related functionality is
    executed in the second stage.
    """

    stdout_handle = sys.stdout  # Unless redefined - do not capture stdout
    stderr_handle = sys.stderr  # Unless redefined - do not capture stderr


    if defer_run:
        cl = get_deferred_cl(info)
    else:
        cl = info['cl']
        if info['stdout_file']:
            lg.debug('capturing stdout in %s', info['stdout_file'])
            stdout_handle = open(info['stdout_file'], 'w')
        if info['stderr_file']:
            lg.debug('capturing stderr in %s', info['stderr_file'])
            stderr_handle = open(info['stderr_file'], 'w')

    info['start'] = datetime.utcnow()

    if defer_run:
        P = sp.Popen(cl, shell=True)
        info['pid'] = P.pid
        info['submitted'] = datetime.utcnow()
    else:
        P = sp.Popen(" ".join(cl), shell=True, stdout=sp.PIPE, stderr=sp.PIPE)

        info['pid'] = P.pid
        try:
            psu = psutil.Process(P.pid)
            info['psutil_process'] = psu
            store_process_info(info)
        except psutil.NoSuchProcess, psutil.AccessDenied:
            #job may have already finished - ignore
            pass
        
        stdout_dq = deque(maxlen=100)
        stderr_dq = deque(maxlen=100)

        stdout_len = 0
        stderr_len = 0

        stdout_sha = hashlib.sha1()
        stderr_sha = hashlib.sha1()
        
        while True:
            pc = P.poll()

            if not pc is None:
                break

            #read_proc_status(td)
            time.sleep(0.2)
            store_process_info(info)
            try:
                stdout_len += streamer(P.stdout, stdout_handle, stdout_dq, stdout_sha)
                stderr_len += streamer(P.stderr, stderr_handle, stderr_dq, stderr_sha)
            except IOError as e:
                #it appears as if one of the pipes has failed.
                if e.errno == 32:
                    errors.append("Broken Pipe")
                    #broken pipe - no problem.
                else:
                    message('err', str(dir(e)))
                    errors.append("IOError: " + str(e))
                break

        if stdout_len > 0:
            info['stdout_sha1'] = stdout_sha.hexdigest()
        if stderr_len > 0:
            info['stderr_sha1'] = stderr_sha.hexdigest()
        info['stop'] = datetime.utcnow()
        info['runtime'] = (info['stop'] - info['start']).total_seconds()
        if 'psutil_process' in info:
            del info['psutil_process']
        info['returncode'] = P.returncode
        info['stdout_len'] = stdout_len
        info['stderr_len'] = stderr_len


class BasicExecutor(object):

    def __init__(self, app):
        lg.debug("Starting executor")
        self.app = app

        try:
            self.threads =  self.app.args.threads
        except AttributeError:
            self.threads = 1
            
        if self.threads < 2:
            self.simple = True
        else:
            self.simple = False
            self.pool = ThreadPool(self.threads)
            lg.debug("using a threadpool with %d threads", self.threads)


    def fire(self, info):
        lg.debug("start execution")

        if self.simple:
            simple_runner(info)
        else:
            self.pool.apply_async(simple_runner, [info,], {'defer_run': False})

    def finish(self):
        if not self.simple:
            lg.info('waiting for the threads to finish')
            self.pool.close()
            self.pool.join()
            lg.debug('finished waiting for threads to finish')


class DummyExecutor(BasicExecutor):

    def fire(self, info):
        lg.debug("start dummy execution")
        cl = copy.copy(info['cl'])

        if info['stdout_file']:
            cl.extend(['>', info['stdout_file']])
        if info['stderr_file']:
            cl.extend(['2>', info['stderr_file']])

        lg.debug("  cl: %s", cl)
        print " ".join(cl)
        info['mode'] = 'synchronous'


conf = leip.get_config('kea')
conf['executors.simple'] = BasicExecutor
conf['executors.dummy'] = DummyExecutor




#!/usr/bin/env python
# -*- coding: utf-8 -*-

# CAVEAT UTILITOR
#
# This file was automatically generated by Grako.
#
#    https://pypi.python.org/pypi/grako/
#
# Any changes you make to it will be overwritten the next time
# the file is generated.


from __future__ import print_function, division, absolute_import, unicode_literals
from grako.parsing import graken, Parser


__version__ = (2014, 12, 11, 18, 8, 46, 3)

__all__ = [
    'cramtoolsParser',
    'cramtoolsSemantics',
    'main'
]


class cramtoolsParser(Parser):
    def __init__(self, whitespace=None, nameguard=True, **kwargs):
        super(cramtoolsParser, self).__init__(
            whitespace=whitespace,
            nameguard=nameguard,
            **kwargs
        )

    @graken()
    def _start_(self):
        with self._group():
            self._token('cramtools')

            def block0():
                with self._choice():
                    with self._option():
                        with self._group():
                            self._token('cram')

                            def block1():
                                with self._choice():
                                    with self._option():
                                        with self._group():
                                            self._token('--capture-tags')
                                            self._pattern(r'.*')
                                    with self._option():
                                        self._token('--capture-all-tags')
                                    with self._option():
                                        self._token('--encrypt')
                                    with self._option():
                                        with self._group():
                                            self._token('-I')
                                            self._pattern(r'.*.bam')
                                            self.ast['input_bam'] = self.last_node
                                    with self._option():
                                        with self._group():
                                            self._token('-O')
                                            self._pattern(r'.*.cram')
                                            self.ast['output_cram'] = self.last_node
                                    with self._option():
                                        with self._group():
                                            self._token('-R')
                                            self._pattern(r'.*')
                                            self.ast['input_reference_genome'] = self.last_node
                                    self._error('expecting one of: --capture-all-tags --capture-tags --encrypt -I -O -R')
                            self._closure(block1)
                    with self._option():
                        with self._group():
                            self._token('bam')

                            def block6():
                                with self._choice():
                                    with self._option():
                                        with self._group():
                                            with self._choice():
                                                with self._option():
                                                    self._token('--password')
                                                with self._option():
                                                    self._token('-p')
                                                self._error('expecting one of: --password -p')
                                    with self._option():
                                        with self._group():
                                            with self._group():
                                                with self._choice():
                                                    with self._option():
                                                        self._token('-f')
                                                    with self._option():
                                                        self._token('-required-flags')
                                                    self._error('expecting one of: -f -required-flags')
                                            self._token('re#\\w+')
                                    self._error('expecting one of: --password -f -p -required-flags')
                            self._closure(block6)
                    with self._option():
                        pass
                    self._error('expecting one of: bam cram')
            self._closure(block0)
            self._check_eof()

        self.ast._define(
            ['input_bam', 'output_cram', 'input_reference_genome'],
            []
        )


class cramtoolsSemantics(object):
    def start(self, ast):
        return ast


def main(filename, startrule, trace=False, whitespace=None, nameguard=None):
    import json
    with open(filename) as f:
        text = f.read()
    parser = cramtoolsParser(parseinfo=False)
    ast = parser.parse(
        text,
        startrule,
        filename=filename,
        trace=trace,
        whitespace=whitespace,
        nameguard=nameguard)
    print('AST:')
    print(ast)
    print()
    print('JSON:')
    print(json.dumps(ast, indent=2))
    print()

if __name__ == '__main__':
    import argparse
    import string
    import sys

    class ListRules(argparse.Action):
        def __call__(self, parser, namespace, values, option_string):
            print('Rules:')
            for r in cramtoolsParser.rule_list():
                print(r)
            print()
            sys.exit(0)

    parser = argparse.ArgumentParser(description="Simple parser for cramtools.")
    parser.add_argument('-l', '--list', action=ListRules, nargs=0,
                        help="list all rules and exit")
    parser.add_argument('-n', '--no-nameguard', action='store_true',
                        dest='no_nameguard',
                        help="disable the 'nameguard' feature")
    parser.add_argument('-t', '--trace', action='store_true',
                        help="output trace information")
    parser.add_argument('-w', '--whitespace', type=str, default=string.whitespace,
                        help="whitespace specification")
    parser.add_argument('file', metavar="FILE", help="the input file to parse")
    parser.add_argument('startrule', metavar="STARTRULE",
                        help="the start rule for parsing")
    args = parser.parse_args()

    main(
        args.file,
        args.startrule,
        trace=args.trace,
        whitespace=args.whitespace,
        nameguard=not args.no_nameguard
    )




"""
Map arguments

input:

    {name~*}

    the * is treated as a glob and can be part of a filename
    name is used to expand the capture part later.

    {name=[1-5]}

    name is replaced by the range of

    {name=a,b,c,d}

"""

import copy
from datetime import datetime
import glob
import logging
import itertools
import re
import shlex
import sys
from collections import OrderedDict

lg = logging.getLogger(__name__)

RE_FIND_MAPINPUT = re.compile(r'{([a-zA-Z_][a-zA-Z0-9_]*)([\~\=])([^}]+)}')


def map_range_expand(map_info, cl, pipes):
    """
    Convert a map range to a list of items:

    first - range conversion:
      - a,b,c -> ['a', 'b', 'c']
      - 5:10 -> [5,6,7,8,9]
      - 5:10:2 -> [5,7,9]
    then, merge with the command line item in which this was embedded (if any)

    so: test.{a=5:10:2}.in becomes: ['test.5.in', 'test.7.in', 'test.9.in']

    """

    mappat_range_3 = re.match(r'([0-9]+):([0-9]+):([0-9]+)',
                              map_info['pattern'])
    mappat_range_2 = re.match(r'([0-9]+):([0-9]+)',
                              map_info['pattern'])

    thisarg = map_info['arg']
    iterstring = map_info['iterstring']
    argi = cl.index(thisarg)

    if mappat_range_3:
        start, stop, step = mappat_range_3.groups()
        map_items = map(str, range(int(start), int(stop), int(step)))
    elif mappat_range_2:
        start, stop = mappat_range_2.groups()
        map_items = map(str, range(int(start), int(stop)))
    elif ',' in map_info['pattern']:
        map_items = [x.strip() for x in map_info['pattern'].split(',')]
    else:
        lg.critical("Can not parse range: %s", map_info['pattern'])
        exit(-1)

        
    substart = thisarg.index(iterstring)
    subtail = substart + len(iterstring) 
    for g in map_items:
        newcl = copy.copy(cl)
        argrep = thisarg[:substart] + str(g) +  thisarg[subtail:]
        newcl[argi] = argrep
        for j, rarg in enumerate(newcl):
            newcl[j] = map_info['rep_from'].sub(g, rarg)


        new_pipes = []
        for p in pipes:
            if p is None:
                new_pipes.append(None)
            else:
                new_pipes.append(map_info['rep_from'].sub(g, p))

        yield newcl, new_pipes


def map_glob_expand(map_info, cl, pipes):
    thisarg = map_info['arg']
    argi = cl.index(thisarg)
    iterpat = map_info['pattern']
    iterstring = map_info['iterstring']

    globpattern = RE_FIND_MAPINPUT.sub(iterpat, thisarg)
    globhits = glob.glob(globpattern)

    if len(globhits) == 0:
        lg.critical("No files matching pattern: '%s' found", globpattern)
        exit(-1)

    substart = thisarg.index(iterstring)
    subtail = len(thisarg) - (substart + len(iterstring))
    
    for g in globhits:
        newcl = copy.copy(cl)
        newcl[argi] = g
        subrep = g[substart:]
        if subtail > 0:
            subrep = subrep[:-subtail]
        for j, rarg in enumerate(newcl):
            newcl[j] = map_info['rep_from'].sub(subrep, rarg)

        new_pipes = []
        for p in pipes:
            if p is None:
                new_pipes.append(None)
            else:
                new_pipes.append(map_info['rep_from'].sub(g, p))

        yield newcl, new_pipes


def map_iter(map_info):
    for i in map_info['items']:
        map_clean = copy.copy(map_info)
        map_clean['item'] = i
        yield map_clean


def apply_map_info_to_cl(newcl, map_info):
    item = map_info['item']
    for i, arg in enumerate(newcl):
        map_to_re = map_info['re_replace']
        map_fr_re = map_info['re_from']
        if map_fr_re.search(arg):
            newcl[i] = map_fr_re.sub(item, arg)
        elif map_to_re.search(arg):
            newcl[i] = map_to_re.sub(item, arg)

    return newcl

def get_uid(app, runno=None):
    if runno is None:
        no = ''
    else:
        no = '{}.'.format(runno)
        
    if app.args.uid:
        uid = '{}.{}{}'.format(app.name, no, app.args.uid)
    else:
        now = datetime.now()
        # YYYY-MM-DDThh:mm:ss.sTZD (eg 1997-07-16T19:20:30.45+01:00)
        uid = '{}.{}{}'.format(app.name, no, now.strftime("%Y-%m-%dT%H:%M:%S"))
    return(uid)
    
    
def basic_command_line_generator(app):
    """
    Most basic command line generator possible
    """
    
    info = OrderedDict()
    
    pipes = [app.args.stdout, app.args.stderr]

    cl = app.cl

    cljoin = " ".join(cl)

    #check if there are iterable arguments in here
    mapcount = 0
    mapins = RE_FIND_MAPINPUT.search(cljoin)

    # no map definitions found - then simply return the cl & execute
    if mapins is None:
        info['cl'] = cl
        info['run_no'] = 0
        info['run_uid'] = get_uid(app)
        info['stdout_file'] = pipes[0]
        info['stderr_file'] = pipes[1]
        yield info
        return

    info['template_cl'] = cl
    #define iterators for each of the definitions

    #lg.setLevel(logging.DEBUG)

    lg.debug("iterables found in: %s", cljoin)
    
    def expander(cl, pipes):
        
        for i, arg in enumerate(cl):
            mima = RE_FIND_MAPINPUT.search(arg)
            if mima:
                break
        else:
            yield cl, pipes
            return

        map_info = {}
        map_info['re_search'] = mima
        map_info['name'], map_info['operator'], map_info['pattern'] = \
                mima.groups()

        map_info['iterstring'] = mima.group(0)
        map_info['arg'] = arg
        

        lg.debug("iterable name: {name} operator: {operator} pattern: {pattern}".format(**map_info))
        
        map_info['re_from'] = re.compile(r'({' + map_info['name']
                                         + r'[\~\=][^}]*})')
        map_info['re_replace'] = re.compile(r'({' + map_info['name'] + r'})')
        map_info['rep_from'] = re.compile(r'({' + map_info['name'] + '})')
        
        map_info['start'] = mima.start()
        map_info['tail'] = len(arg) - mima.end()

        if map_info['operator'] == '~':
            expand_function = map_glob_expand
        elif map_info['operator'] == '=':
            expand_function = map_range_expand
        
        for ncl, pipes in  expand_function(map_info, cl, pipes):
            for nncl, pipes in expander(ncl, pipes):
                    yield nncl, pipes

    no = 0
    for newcl, pipes in  expander(cl, pipes):
        no += 1
        
        newinfo = copy.copy(info)
        newinfo['run_uid'] = get_uid(app, no)
        newinfo['cl'] = newcl
        newinfo['run_no'] = no
        newinfo['stdout_file'] = pipes[0]
        newinfo['stderr_file'] = pipes[1]
        yield newinfo



"""
Parse command lines using grako generated parsers
"""

import logging
import imp

from grako.exceptions import FailedParse

import leip

lg = logging.getLogger(__name__)
#lg.setLevel(logging.DEBUG)

PMODS = {}


@leip.hook('pre_fire')
def parse_commandline(app, jinf):
    
    global PMODS

    lg.debug("pre_fire start for command: %s", app.name)
        
    if app.name in PMODS:
        parser = PMODS[app.name]
    else:
        #find module
        for ml in app.conf['plugin.cl_parser.parse_locations']:
            modname = '{}.{}'.format(ml, app.name)
            try:
                module = __import__(modname, fromlist=[app.name])
                break
            except ImportError:
                #no module found - ignore
                continue
        else:
            lg.debug("unable to import a parser module for %s", app.name)
            return

        parserclass = getattr(module, '{}Parser'.format(app.name))
        parser = parserclass()

        PMODS[app.name] = parser

        
    cl = " ".join([app.name] + jinf['cl'][1:])
    lg.debug("parsing cl: %s", cl)
    
    try:
        p = parser.parse(cl, rule_name='start')
    except FailedParse:
        lg.info("Cannot parse command line: %s", cl)
        return

    ast = p[1]
    if len(ast) == 0:
        return

    jinf['ast'] = jinf.get('ast', {})
    for k in ast:
        v = ast[k]
        if not v: continue
#        if not isinstance(v, list):
#            v = [v]
        jinf['ast'][k] = v

    




from collections import OrderedDict
import logging

import arrow
from mad2 import madfile
import leip
from lockfile import FileLock

lg = logging.getLogger(__name__)

def to_str(s):
    if isinstance(s, madfile.MadFile):
        if 'sha1sum' in s:
            return '{} (sha1: {})'.format(s['inputfile'], s['sha1sum'])
        else:
            return '{}'.format(s['inputfile'])
    else:
        return str(s)


@leip.hook('pre_argparse')
def logger_arg_define(app):
    app.parser.add_argument('-S', '--report_screen', action='store_true')
    app.parser.add_argument('-Y', '--report_yaml', action='store_true')

@leip.hook('post_fire')
def log_screen(app, jinf):
    
    if app.args.report_yaml:
        import yaml
        fn = '{}.report.yaml'.format(jinf['run_uid'])
        with open(fn, 'w') as F:
            yaml.dump(jinf, F, default_flow_style=False)
        
    
    if not app.args.report_screen:
        return

    def dictprint(d, pref=""):
        mxkyln = max([len(x) for x in d.keys()])
        fs = pref + '{:<' + str(mxkyln) + '} : {}'
        for k in sorted(d.keys()):        
            v = d[k]
            if k in ['cl', 'template_cl']:
                v = " ".join(v)
                
            if v is None: continue
            if v == "": continue
            if not isinstance(v, dict):
                print fs.format(k, v)
            else:
                dictprint(v, '{}.'.format(k))
                
    print '--KEA-REPORT' + '-' * 68
    dictprint(jinf)
    print '-' * 80
    
@leip.hook('post_run')
def log_cl(app):
    all_jinf = app.all_jinf
    try:
        with FileLock('kea.log'):
            for i, info in enumerate(all_jinf):
                with open('kea.log', 'a') as F:
                    F.write("-" * 80 + "\n")
                    for i in info:
                        F.write("{}: ".format(i))
                        val = info[i]
                        if i == 'cl':
                            F.write(" ".join(val) + "\n")
                        elif isinstance(val, list):
                            F.write("\n")
                            for lv in val:
                                F.write(' - {}\n'.format(to_str(lv)))
                        else:
                            F.write(" {}\n".format(to_str(val)))

    except:
        lg.warning("Cannot write to log file")



#!/usr/bin/env python

import re
import sys
from collections import OrderedDict

import yaml

psplit = re.compile('#[-]')

clfile = sys.argv[1]

def interject(r, inter = '|'):
    for i, e in enumerate(r):
        if i > 0:
            yield inter
        yield e

def process_endnode(n):
    if not n.strip():
        return "()"
    if n.startswith("sp#"):
        return n[3:]
    if re.search(r'^[a-zA-Z_]+!!', n):
        name, nnode = n.split("!!", 1)
        return "{}:{}".format(name, process_node(nnode))
    if n.startswith("/") and n.endswith("/"):
        return n + " /\s*/"
    else:
        return  '"{}" /\s*/'.format(n)

def process_listnode(n):
    return ["("] + map(process_node, n), [")"]

def process_oneofnode(n):
    return ["("] + list(interject(map(process_node, n), '|')) + [")"]

def process_namednode(n):
    name = n['name']
    value = n['value']
    return "{}:{}".format(name, process_node(value))


def process_zeroormore(n):
    return ["{"] + list(interject(map(process_node, n), '|')) + ["}"]

def process_dictnode(n):
    assert len(n) == 1
    k, v = n.items()[0]
    if k == 'oneof':
        return process_oneofnode(v)
    elif k == 'zeroormore':
        return process_zeroormore(v)
    elif k == 'named':
        return process_namednode(v)
    else:
        print "\n\n\nPROBLEM\n"
        print(k)
        print(v)
        exit()
 

def process_node(n=""):
    if isinstance(n, str):
        return process_endnode(n)
    elif isinstance(n, list):
        return process_listnode(n)
    elif isinstance(n, dict):
        return process_dictnode(n)

def listprint(lst, indent=""):
    for i in lst:
        if isinstance(i, str):
            print indent + i
        elif isinstance(i, list) or isinstance(i, tuple):
            listprint(i, indent+"  ")

def blockparser(F):
    
    for line in F:
    
with open(sys.argv[1]) as F:
    rv = ["start = "]
    for block in blockparser(F):
        print(block)
    
#    listprint(process_node(data))
#    print(" ; ")




import leip
import logging
import subprocess as sp

import jinja2

lg = logging.getLogger(__name__)


@leip.hook('pre_argparse')
def define_args(app):
    maildef = app.conf['plugin.mail.default']
    if maildef:
        app.parser.add_argument('-M', '--no_mail', help='do NOT send an email report', action='store_true')
    else:
        app.parser.add_argument('-m', '--mail', help='send an email report', action='store_true')
    mailsuc = app.conf['plugin.mail.on_success']
    if mailsuc:
        app.parser.add_argument('-U', '--no_mail_on_success', dest='mail_on_success',
                                help='do NOT send an email report on successful execution',
                                action='store_false')
    else:
        app.parser.add_argument('-u', '--mail_on_success', help='send an email report on ' +
                                'successful execution', action='store_true')

        
HTML_MESSAGE = """MIME-Version: 1.0
Content-Type: text/html;
Subject: 
{%- if success %} Kea/Ok: {{cl50}} 
{%- else %} Kea/Fail: {{ cl50 }} 
{%- endif %}


{% if success -%}
<h2 style='color:green'>Kea/Job OK</h2>
{%- else -%}
<h2 style='color:red'>Kea/Job FAIL</h2>
{%- endif %}

{% if all_jinf|length > 1 -%}
<h3>Template Command Line</h3>
<pre>
  {{ clj }}
</pre>

<h3>Actual command lines</h3>
{%- else -%}
<h3>Command line</h3>{% endif %}

<pre>{% for jinf in all_jinf -%}
{{ jinf.run_no }}.  {{ jinf['cl']|join(" ") }}

{% endfor %}
</pre>
{%- if all_jinf|length > 1 %}
<h3>Run stats</h3>
 
<table>
<tr>
  <th>No.</th>
{% if not success %}
  <th>rc</th>
{% endif %}
  <th>Run time</th>
  <th>Cpu time</th>
  <th>Peak mem</th>
  <th>Stdout</th>
  <th>Stderr</th>
</tr>
{% for jinf in all_jinf %}
<tr>
  <td><b>{{ jinf.run_no }}</b></td>
  {%- if not success %}{% if jinf.returncode == 0 %}
  <td><span style='color:green'>0</span></td>
  {% else %}
  <td><span style='color:red;weight:bold;'>{{jinf.returncode}}</span></td>
  {%- endif %}{% endif %}
  <td>{{ jinf.runtime }}</td>
  <td>{{ "%.3f"|format(jinf.ps_cputime_user) }}</td>
  <td>{{ "%d"|format(jinf.ps_meminfo_max_rss) }}</td>
  <td>{{ jinf.stdout_len }}</td>
  <td>{{ jinf.stderr_len }}</td>
</tr>
{%- endfor %}
</table>
{% endif %}

{% for jinf in all_jinf %}
<h3>Full run report {% if all_jinf|length > 1 %} ({{ jinf.run_no }}) {% endif %}</h3>
<table>
{% for k in jinf %}
<tr {% if loop.cycle(False, True) -%}style="background-color: #EEEEEE;"
    {%- endif %}><th style="text-align: left;">{{k}}</th>
    {%- if k == "cl" %}
      <span style="font-family:Lucida Console,Bitstream Vera Sans Mono,Courier New,monospace;">{{ " ".join(jinf.cl) }}</span>
    {%- else %}
    <td>{{ jinf[k] }}</td>
    {%- endif %}
</tr>
{%- endfor %}
</table>

{% endfor %}

"""

@leip.hook('finish')
def mail(app):

    if not hasattr(app, 'all_jinf'):
        return
    
    maildef = app.conf['plugin.mail.default']
    mailsuc = app.conf['plugin.mail.on_success']

    #did all jobs finish successfully?
    success = True
    
    for jinf in app.all_jinf:
        if jinf['returncode'] != 0:
            success = False
            break

    data = {}
    data['success'] = success
    data['all_jinf'] = app.all_jinf
    data['app'] = app
    data['clj50'] = " ".join(app.cl)[:50]
    data['clj'] = " ".join(app.cl)

    
    if success and not app.args.mail_on_success:
        #all is well - do not send an email
        return

    # mkea conf set plugin.mail.recipient 'mark.fiers@cme.vib-kuleuven.be'
    mailto = app.conf['plugin.mail.recipient']
    
    if not mailto:
        lg.warning("No mail recipient defined - cannot send mail")
        return
    

    message = jinja2.Template(HTML_MESSAGE)
    message = message.render(data)

    p = sp.Popen("sendmail %s" % mailto, stdin=sp.PIPE, shell=True)
    p.communicate(message.encode('utf-8'))


    exit(p.returncode)




import copy
import logging
import os

from uuid import uuid4
from jinja2 import Template
import arrow

import leip

from kea.executor import BasicExecutor, get_deferred_cl


lg = logging.getLogger(__name__)


@leip.hook('pre_argparse')
def prep_sge_exec(app):


    app.parser.add_argument('--pbs_nodes',
                                  help='No nodes requested (default=jobs '
                                  + 'submitted)', type=int)
    app.parser.add_argument('--pbs_ppn',
                                  help='No ppn requested (default=cl per '
                                  + 'job)', type=int)
    app.kea_argparse.add_argument('--pbs_account', '--pbs_account',
                                  help='Account requested (default none)')




PBS_SUBMIT_SCRIPT_HEADER = """#!/bin/bash
#PBS -N {{appname}}.{{uuid}}
#PBS -e {{ cwd }}/pbs/{{appname}}.{{uuid}}.$PBS_JOBID.err
#PBS -o {{ cwd }}/pbs/{{appname}}.{{uuid}}.$PBS_JOBID.out
#PBS -l nodes={{nodes}}:ppn={{ppn}}

{% if account -%}
  #PBS -A {{ account }}{% endif %}
{% if walltime -%}
  #PBS -l walltime={{ walltime }}{% endif %}

set -v

"""


class PbsExecutor(BasicExecutor):

    def __init__(self, app):

        super(PbsExecutor, self).__init__(app)
        if not os.path.exists('./pbs'):
            os.makedirs('./pbs')

        self.buffer = []
        self.cl_per_job =  self.app.kea_args.threads
        if self.cl_per_job == -1:
            self.cl_per_job = 1
        self.batch = 0
        self.clno = 0


    def submit_to_pbs(self):
        uuid = str(uuid4())[:8]
        #write pbs script
        pbs_script = os.path.join(
                  'pbs',
                  '{}.{}.pbs'.format(self.app.conf['appname'], uuid))

        template = Template(PBS_SUBMIT_SCRIPT_HEADER)

        data = copy.copy(self.app.conf['pbs'])
        data['appname'] = self.app.conf['appname']
        data['cwd'] = os.getcwd()
        data['uuid'] = uuid

        lg.debug("submit to pbs with uuid %s", uuid)
        for info in self.buffer:

            #for logging.
            info['mode'] = 'asynchronous'
            info['submitted'] = arrow.utcnow()
            info['pbs_uuid'] = uuid
            info['pbs_script_file'] = pbs_script

            if not self.app.kea_args.pbs_nodes is None:
                data['nodes'] = self.app.kea_args.pbs_nodes
            elif not data['nodes']:
                data['nodes'] = 1

            if self.app.kea_args.pbs_ppn:
                data['ppn'] = self.app.kea_args.pbs_ppn
            elif not data['ppn']:
                data['ppn'] = self.cl_per_job

            with open(pbs_script, 'w') as F:
                data['name']
                F.write(template.render(**data))
                for info in self.buffer:
                    F.write("( " + " ".join(get_deferred_cl(info)))
                    F.write(" ) & \n")
                F.write("wait\n")
                F.write('echo "done"\n')

            self.clno += 1

        #fire & forget the pbs job
        pbs_cl = ['qsub', pbs_script]
        print " ".join(pbs_cl)
        self.buffer = []
        self.batch += 1


    def fire(self, info):
        self.buffer.append(info)
        if len(self.buffer) >= self.cl_per_job:
            lg.info("submitting pbs job. No commands: %d", len(self.buffer))
            self.submit_to_pbs()

    def finish(self):
        if len(self.buffer) > 0:
            lg.info("submitting pbs job. No commands: %d", len(self.buffer))
            self.submit_to_pbs()


conf = leip.get_config('kea')
conf['executors.pbs'] = PbsExecutor

